from typing import Optional

from .. import LoginID
from ..core import LoginIDError


class _CredentialManagement():
    """
    Credential management class.
    """

    # Credentials
    def get_credentials(self: LoginID,
                        status: Optional[str] = None,
                        user_id: Optional[str] = None,
                        username: Optional[str] = None) -> dict:
        """Get an exhaustive list of credentials for a given user

        Args:
            status (str, optional): The status of the credentials to get. Defaults to None for credentials with any status.
            user_id (str, optional): User ID of the end user who you would like to get the list of credentials for. Defaults to None.
            username (str, optional): Username of the end user, must be present if `user_id` is not. Defaults to None.

        Returns:
            dict: User's credentials if no error

        Raises:
            LoginIDError: If there is an error
        """
        url = f"/credentials/list"

        payload = {"client_id": self._client_id}
        self._update_payload(payload, user_id, username)

        if status:
            payload["status"] = status

        return self._request('post', url, "credentials.list", payload=payload)

    def rename_credential(self: LoginID,
                          cred_id: str,
                          updated_name: str,
                          user_id: Optional[str] = None,
                          username: Optional[str] = None) -> dict:
        """Rename a credential of a user

        Args:
            cred_id (str): The id of the credential to be renamed.
            updated_name (str): The new name.
            user_id (str, optional): The id of the user to rename the credential for. Defaults to None.
            username (str, optional): The username, must be present if `user_id` is not. Defaults to None.

        Returns:
            dict: The renamed credential's detail if no error

        Raises:
            LoginIDError: If there is an error
        """

        url = "/credentials/rename"
        payload = {
            "client_id": self._client_id,
            "credential": {
                "uuid": cred_id,
                "name": updated_name
            }
        }

        self._update_payload(payload, user_id, username)

        return self._request('post', url, "credentials.rename", payload)

    def revoke_credential(self: LoginID,
                          cred_id: str,
                          user_id: Optional[str] = None,
                          username: Optional[str] = None) -> dict:
        """Revoke an existing credential from a user

        Args:
            cred_id (str): The credential id to be revoked
            user_id (str, optional): The user id to extract the credential from. Defaults to None.
            username (str, optional): The username, must be present if `user_id` is not. Defaults to None.

        Returns:
            dict: The revoked credential's detail

        Raises:
            LoginIDError: If there is an error
        """

        url = "/credentials/revoke"
        payload = {
            "client_id": self._client_id,
            "credential": {
                "uuid": cred_id
            }
        }
        self._update_payload(payload, user_id, username)

        return self._request('post', url, "credentials.revoke", payload)

    def force_fido2_credential_init(self: LoginID,
                                    user_id: str,
                                    display_name: Optional[str] = None,
                                    roaming_authenticator: bool = False) -> dict:
        """Initialize adding a FIDO2 credential without pre-generated authorization code
        Args:
            user_id (str): ID of the user to add new credential to.
            display_name (str, optional): A human-palatable name for the user account, intended only for display. Defaults to None.
            roaming_authenticator (bool, optional): Whether a security key is allowed. Defaults to False.
        Returns:
            dict: The attestation payload for the credentials if no errors

        Raises:
            LoginIDError: If there is an error
        """
        url = '/credentials/fido2/init/force'

        payload = {
            "client_id": self._client_id,
            "user_id": user_id
        }

        options = {}
        if display_name:
            options['display_name'] = display_name
        if roaming_authenticator:
            options['roaming_authenticator'] = roaming_authenticator
        if options:
            payload['options'] = options

        return self._request('post', url, "credentials.force_add", payload)

    def add_doc_scan_credential_init(self: LoginID,
                                     user_id: Optional[str] = None,
                                     username: Optional[str] = None,
                                     credential_name: Optional[str] = None) -> dict:
        """Initialize the adding of a document scan credential

        Args:
            user_id (str optional): ID of the user to add new credential to. Defaults to None.
            username (str, optional): the username to add new credential to, must be present if `user_id` is not. Defaults to None.
            credential_name (str, optional): the name of the new credential. Defaults to None.

        Returns:
            dict: dictionary of form::

                {
                    'credential_id': 'uuid-string',
                    'iframe_url': 'https://ipsidy.com/url',
                }

        Raises:
            LoginIDError: If there is an error
        """
        url = '/credentials/authid/init'

        payload = {
            "client_id": self._client_id,
            "user_id": user_id,
            "username": username
        }
        self._update_payload(payload, user_id, username)

        if credential_name:
            payload['options'] = {
                "credential_name": credential_name
            }

        return self._request('post', url, "credentials.force_add", payload)

    def add_doc_scan_credential_complete(self: LoginID, credential_id: str, activate_credential: bool,
                                         user_id: Optional[str] = None, username: Optional[str] = None,
                                         ) -> dict:
        """Complete the adding of a document scan credential

        Args:
            credential_id (str): The credential id
            active_credential (bool): Whether to activate the credential
            user_id (str, optional): ID of the user to add new credential to. Defaults to None.
            username (str, optional): The username, must be present if `user_id` is not. Defaults to None.

        Returns:
            dict: The Fido2 attestation response for the added credential if no errors

        Raises:
            LoginIDError: If there is an error
        """

        url = '/credentials/authid/complete'
        payload = {
            "client_id": self._client_id,
            "credential_uuid": credential_id,
            "activate_credential": activate_credential,
        }
        self._update_payload(payload, user_id, username)

        response = self._request('post', url, "credentials.force_add", payload)

        if activate_credential:
            token = response.get('jwt', '')
            if not self.verify_token(token):
                raise LoginIDError(401, "invalid_token", "Invalid token")

        return response

    def evaluate_doc_scan_credential(self: LoginID, credential_id: str, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        """Retrieve document scan and evaluate the credential

        Args:
            credential_id (str): The credential id
            user_id (str, optional): ID of the user to add new credential to. Defaults to None.
            username (str, optional): The username, must be present if `user_id` is not. Defaults to None.

        Returns:
            dict: a dictionary::

                {
                    "request_url": "https://example.com/authid/evaluate",
                    "token_type": "Bearer",
                    "auth_token": "auth-token-string"
                }

        Raises:
            LoginIDError: If there is an error
        """

        url = '/credentials/authid/evaluate'
        payload = {
            "client_id": self._client_id,
            "credential_uuid": credential_id,
        }
        self._update_payload(payload, user_id, username)

        return self._request('post', url, "credentials.retrieve_sensitive", payload)

    def generate_recovery_code(self: LoginID,
                               user_id: Optional[str] = None,
                               username: Optional[str] = None) -> dict:
        """Generate a recovery code

        Args:
            user_id (str, optional): the id of the user to add recovery code. Defaults to None.
            username (str, optional): the username of the user to add recovery code, must be present if `user_id` is not. Defaults to None.
        Returns:
            dict: raise a LoginIDError if there is an error, else a dictionary::

                {
                    "code" : "paraphrase recovery code",
                    "created_at": "2022-02-01T21:17:52.184Z"
                }
        """

        url = '/credentials/recovery-code'
        payload = {
            "client_id": self._client_id,
        }
        self._update_payload(payload, user_id, username)

        return self._request('post', url, "credentials.add_recovery", payload)

    def add_publickey_credential(self: LoginID, publickey: str, publickey_alg: Optional[str] = "ES256",
                                 user_id: Optional[str] = None, username: Optional[str] = None,
                                 credential_name: Optional[str] = None) -> dict:
        """Add a public key as a new credential

        Args:
            publickey (str): the public key in PEM format
            user_id (str, optional): the id of the user to add credential. Defaults to None.
            username (str, optional): the username, must be present if `user_id` isn't. Defaults to None.
            credential_name (str, optional): the name of the credential. Defaults to None.

        Returns:
            dict: raise a LoginIDError if there is an error, else a dictionary
        """
        if not (user_id or username):
            raise LoginIDError(400, "bad_request",
                               "`user_id` or `username` must be present")

        url = '/credentials/publickey'
        payload = {
            "client_id": self._client_id,
            "publickey_alg": publickey_alg,
            "publickey": publickey
        }
        self._update_payload(payload, user_id, username)

        if credential_name:
            payload['options'] = {
                "credential_name": credential_name
            }

        return self._request('post', url, "credentials.force_add", payload)



from typing import Tuple, Union, Dict
class LoginIDError(Exception):
	"""Base class for exceptions in this module."""
	def __init__(self, status_code:int, error_code: str, message: str):
		self.status_code = status_code
		self.error_code = error_code
		self.message = message

		super().__init__(self.message)

	def __str__(self):
		return f"HTTP status code: {self.status_code}, error code: {self.error_code}, error message: {self.message}"


from typing import Optional
from . import _LoginIdClient, LoginIDError


class _Auth:

    ###########################################################################
    #### Registration                                                      ####
    ###########################################################################
    def register_fido2_init(self: _LoginIdClient,
                            username: str,
                            display_name: Optional[str] = None,
                            roaming_authenticator: bool = False) -> dict:
        """
        Initiate a FIDO2 registration

        Args:
            username (str): The username to be registered.
            display_name (str, optional): A human-palatable name for the user account, intended only for display. Defaults to None.
            roaming_authenticator (bool, optional): Whether the authenticator is a security key. Defaults to False.
        Returns:
            dict: The Fido2 attestation payload
        """
        url = "/register/fido2/init"

        payload = dict(
            client_id=self._client_id
        )
        self._update_payload(payload, username=username)

        options = {}
        if display_name:
            options["display_name"] = display_name
        if roaming_authenticator:
            options["roaming_authenticator"] = True
        if options:
            payload["options"] = options

        return self._request('post', url, "auth.register", payload=payload)

    def register_fido2_complete(self: _LoginIdClient, username: str,
                                attestation_payload: dict,
                                credential_name: Optional[str] = None) -> dict:
        """Complete a FIDO2 registration

        Args:
            username (str): The username.
            attestation_payload (dict): the attestation payload.
            credential_name (str, optional): The credential name. Defaults to None.
        """

        url = "/register/fido2/complete"

        payload = dict(
            client_id=self._client_id,
            attestation_payload=attestation_payload,
        )
        self._update_payload(payload, username=username)

        if credential_name:
            payload['options'] = {"credential_name": credential_name}

        auth_response = self._request('post', url, "auth.register", payload)
        token = auth_response.get('jwt', '')
        if not self.verify_token(token):
            raise LoginIDError(401, "invalid_token", "Invalid token")

        return auth_response

    def register_password(self: _LoginIdClient, username: str, password: str) -> dict:
        """Register a new user with password

        Args:
            username (str): The username to be registered

        Args:
            username (str): The new username
            password (str): The password

        Returns:
            dict: The registration response
        """
        url = '/register/password'
        payload = {
            "client_id": self._client_id,
            "password": password,
            "password_confirmation": password,
        }
        self._update_payload(payload, username=username)

        auth_response = self._request('post', url, "auth.register", payload)

        token = auth_response.get('jwt', '')
        if not self.verify_token(token):
            raise LoginIDError(401, "invalid_token", "Invalid token")

        return auth_response

    ###########################################################################
    #### Authentication                                                    ####
    ###########################################################################
    def authenticate_password(self: _LoginIdClient,
                              username: str,
                              password: str) -> dict:
        """Authenticate user with password

        Args:
            username (str): The username to be authenticated
            password (str): The password

        Returns:
            dict: The authentication response.
        """
        url = f"/authenticate/password"
        payload = {
            "client_id": self._client_id,
            "password": password,
        }
        self._update_payload(payload, username=username)

        response = self._request('post', url, "auth.login", payload=payload)

        token = response.get('jwt', '')
        if not self.verify_token(token):
            raise LoginIDError(401, "invalid_token", "Invalid token")

        return response

    # Fido2
    def authenticate_fido2_init(self: _LoginIdClient, username: str) -> dict:
        """Initialize authentication process with a FIDO2 credential

        Args:
            username (str): The username to be authenticated.

        Returns:
            dict: a dictionary::

                {
                    "assertion_payload": {
                        "challenge": "string",
                        "rpId": "string",
                        "allowCredentials": [
                            {
                                "id": "string",
                                "type": "fiod2",
                                transports: ["string"]
                            }
                        ],
                    }
                }

        Raises:
            LoginIDError: If there is an error
        """

        url = '/authenticate/fido2/init'
        payload = {
            "client_id": self._client_id,
        }
        self._update_payload(payload, username=username)

        return self._request('post', url, "auth.login", payload)

    def authenticate_fido2_complete(self: _LoginIdClient, username: str, assertion_payload: dict) -> dict:
        """Complete authentication process with a FIDO2 credential

        Args:
            username (str): The username to be authenicated.
            assertion_payload (str): The FIDO2 assertion payload

        Returns:
            dict: The authentication response
        """
        url = '/authenticate/fido2/complete'
        payload = {
            "client_id": self._client_id,
            "assertion_payload": assertion_payload,
        }
        self._update_payload(payload, username=username)

        response = self._request('post', url, payload=payload)

        token = response.get('jwt', '')
        if not self.verify_token(token):
            raise LoginIDError(401, "invalid_token", "Invalid token")

        return response

    # Code
    def authenticate_code_wait(self, username: str, code: str, code_type: str) -> dict:
        """ Wait for a given code

        Args:
            username (str): The username
            code (str): The code associate to the username
            code_type (str): Type of the code

        Returns: response json
        """

        self.validate_code_type(code_type)

        url = f'/authenticate/code/wait'

        payload = {
            "client_id": self._client_id,
            "username": username,
            "authentication_code": {
                "code": code,
                "type": code_type
            }
        }

        json_response = self._request(
            'post', url, token_scope="auth.temporary", payload=payload)

        token = json_response.get('jwt')
        if (token is None) or (not self.verify_token(token)):
            raise LoginIDError(500, "internal_error", "Invalid token")

        return json_response

    def validate_code_type(self, code_type: str) -> None:
        """Check if a code type is valid, raise an error if not

        Args:
            code_type (str): code type

        """
        if code_type not in self.ALLOWED_CODE_TYPES:
            raise LoginIDError(400, 'BAD_REQUEST',
                               f"{code_type} is not a valid code type")

    # document scan
    def authenticate_doc_scan_init(self: _LoginIdClient, username: str, credential_id: Optional[str] = None) -> dict:
        """Initialize authentication process with a document scan

        Args:
            username (str): The username to be authenticated.
            credential_id (str, optional): the credential id to be used in authentication.

        Returns:
            dict: a dictionary::

                {
                    "credential_uuid": "uuid-string",
                    "iframe_url": "https://example.com/authid/authenticate",
                }

        Raises:
            LoginIDError: If there is an error
        """

        url = '/authenticate/authid/init'
        payload = {
            "client_id": self._client_id,
        }
        self._update_payload(payload, username=username)

        if credential_id:
            payload['options'] = {
                'credential_uuid': credential_id
            }

        return self._request('post', url, "auth.login", payload)

    def authenticate_doc_scan_complete(self: _LoginIdClient, username: str, credential_id: str) -> dict:
        """Complete authentication process with a document scan

        Args:
            username (str): The username to be authenticated.
            credential_id (str): The credential id to be used in authentication

        Returns:
            dict: The authentication response upon success.

        Raises:
            LoginIDError: if authentication fails
        """
        url = '/authenticate/authid/complete'
        payload = {
            "client_id": self._client_id,
            "credential_uuid": credential_id,
        }
        self._update_payload(payload, username=username)

        response = self._request('post', url, payload=payload)
        token = response.get('jwt', '')
        if not self.verify_token(token):
            raise LoginIDError(401, "invalid_token", "Invalid token")

        return response

    # publickey
    def authenticate_publickey_init(self: _LoginIdClient,
                                    username: str,
                                    publickey: str,
                                    publickey_alg: Optional[str] = "ES256") -> dict:
        """Initialize an authentication with a public key

        Args:
            username (str): the username to be authenticated.
            publickey (str): the public key in PEM format
            pubickey_alg (str, optional): the algorithm of the public key. Defaults to "ES256".

        Returns:
            dict: raise a LoginIDError if there is an error, else a dictionary::

                {
                    "challenge_id": "uuid-string",
                    "server_nonce": "base64url-encoded-string",
                }
        """

        url = '/authenticate/publickey/init'
        payload = {
            "client_id": self._client_id,
            "publickey_alg": publickey_alg,
            "publickey": publickey
        }
        self._update_payload(payload, username=username)

        return self._request('post', url, token_scope="auth.login", payload=payload)

    def authenticate_publickey_complete(self: _LoginIdClient,
                                        username: str,
                                        challenge_id: str,
                                        assertion: str) -> dict:
        """Complete an authentication with a public key

        Args:
            username (str): the username to be authenticated. Defaults to None.
            challenge_id (str): the challenge id
            assertion (str): the assertion in jwt format

        Returns:
            dict: raise a LoginIDError if there is an error, else a dictionary
        """
        url = '/authenticate/publickey/complete'
        payload = {
            "client_id": self._client_id,
            "challenge_id": challenge_id,
            "assertion": assertion,
        }
        self._update_payload(payload, username=username)

        response = self._request('post', url, payload=payload)

        token = response.get('jwt', '')
        if not self.verify_token(token):
            raise LoginIDError(401, "invalid_token", "Invalid token")

        return response



import requests
import json
import jwt
from .errors import LoginIDError

import random
from datetime import datetime

from typing import Optional

class _LoginIdClient():
    DEFAULT_JWT_ALOGORITHM = 'ES256'
    SUPPORTED_JWT_ALGORITHMS = {DEFAULT_JWT_ALOGORITHM}
    DEFAULT_BASE_URL = "https://directweb.usw1.loginid.io"

    ALLOWED_CODE_TYPES = {'short', 'long', 'phrase'}

    # leeway is in seconds
    JWT_LEEWAY = 60

    def __init__(self, client_id: str, private_key: Optional[str]=None, base_url: Optional[str]=None) -> None:
        self._private_key = private_key
        self._client_id = client_id
        self._base_url = (base_url or self.DEFAULT_BASE_URL).rstrip('/')

    def client_id(self) -> str:
        """Extract the client id

        Returns:
            str: The client id
        """
        return self._client_id

    def _get_utc_epoch(self):
        """
        returns the current UTC epoch in seconds
        """
        return int((datetime.utcnow()-datetime(1970, 1, 1)).total_seconds())

    def _random_string(self, length: int = 16) -> str:
        """
        Generate a random string of given length with alphanumeric characters

        Args:
            length (int): length of the output string

        Returns:
            str: a random string
        """
        possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
        ret = [random.choice(possible) for _ in range(length)]
        return "".join(ret)

    def _get_public_key(self, kid: Optional[str] = None) -> str:
        """Return the server's public key with `kid`

        Args:
            kid (str, optional): the kid included in jwt's header. Defaults to "".

        Returns:
            str: The public key in PEM format
        """
        # ensure kid is not None
        kid = kid or ""

        params = {"kid": kid} if kid else None

        r = requests.get(f'{self._base_url}/certs', params=params)
        if (r.status_code != 200):
            error = r.json()
            raise LoginIDError(
                r.status_code,
                error.get('code', 'UNKNOWN_ERROR'),
                error.get('message', 'Unknown Error')
            )

        # TODO: cache the public key
        return r.text

    def _update_payload(self, payload: dict, user_id: Optional[str] = None, username: Optional[str] = None) -> None:
        """Update payload with user_id and username if present

        Args:
            payload (dict): payload to be updated
            user_id (str, optional): user id to be updated. Defaults to None.
            username (str, optional): username to be updated. Defaults to None.
        """

        # verify that either user_id or username is present
        if not (user_id or username):
            raise LoginIDError(
                400, "bad_request", "either `user_id` or `username` must be present")

        if user_id:
            payload['user_id'] = user_id
        if username:
            payload['username'] = username

    def _request(self, method: str, url: str, token_scope: Optional[str] = None,
                 payload: Optional[dict] = None, params: Optional[dict] = None,
                 headers: Optional[dict] = None, expect=200) -> dict:
        """
        Make a request to the LoginID service
        """
        methods = {
            'post': requests.post,
            'get': requests.get,
            'put': requests.put,
            'delete': requests.delete
        }
        r_method = methods.get(method.lower(), None)
        if r_method is None:
            raise LoginIDError(400, "INVALID_METHOD", "Invalid method")

        if headers is None:
            headers = headers or {
                "Content-type": "application/json",
                "X-Client-ID": self._client_id
            }
            if self._private_key:
                headers['Authorization'] = f"Bearer {self.generate_service_token(token_scope)}"

        _url = f'{self._base_url}{url}'
        if payload is not None:
            r = r_method(_url, data=json.dumps(payload), headers=headers)
        else:
            r = r_method(_url, params=params, headers=headers)

        if (r.status_code != expect):
            try:
                error = r.json()
            except Exception:
                raise LoginIDError(500, 'UNKNOWN_ERROR', r.text)

            raise LoginIDError(
                r.status_code,
                error.get('code', 'UNKNOWN_ERROR'),
                error.get('message', r.text or "Unknown error")
            )
        # only return json if there is a response
        return r.json() if expect != 204 else {}

    def verify_token(self, token: str, username: Optional[str] = None) -> bool:
        """Verify a JWT token returned upon user authorization

        Args:
            token (str): JWT token
            username (str, optional): if given, checks for if `username` matches the `udata` in JWT. Defaults to None.

        Returns:
            bool: `True` if the token is valid, `False` otherwise (including errors)
        """

        headers = jwt.get_unverified_header(token)

        # extract algo, kid from headers, default to None
        algo, kid = headers.get('alg'), headers.get('kid')

        # verify that algo is supported
        if algo not in self.SUPPORTED_JWT_ALGORITHMS:
            raise LoginIDError(400, "invalid_algorithm", f"{algo} is not an allowed algorithm.")

        # obtain public key from LogInID
        public_key = self._get_public_key(kid)

        payload = jwt.decode(
            token, public_key,
            algorithms=self.SUPPORTED_JWT_ALGORITHMS,
            audience=self._client_id,
            leeway=self.JWT_LEEWAY
        )

        if username is not None:
            return username == payload['udata']
        return True

    def generate_service_token(self, scope: str,
                               username: Optional[str] = None, user_id: Optional[str] = None,
                               algo: Optional[str] = None, nonce: Optional[str] = None) -> str:
        """
        Generate a service token

        Args:
            scope (str): the scope of the service
            username (str, optional): the username to-be granted by the token
            user_id (str, optional): the user_id to be granted by the token. If `username` is given, this is ignored
            algo (str, optional): Encryption algorithm, defaults to `"ES256"`
            nonce (str, optional): nonce for the token, randomly generated if not given

        Returns:
            str: the JWT service token if `self._private_key` is set, else the empty string.
        """

        # only generate service token if private key is set
        if not self._private_key:
            return ""

        algo = algo or self.DEFAULT_JWT_ALOGORITHM
        if algo not in self.SUPPORTED_JWT_ALGORITHMS:
            raise LoginIDError(400, "invalid_algorithm", f"{algo} is not an allowed algorithm.")

        payload = {
            "client_id": self._client_id,
            "type": scope,
            "nonce": nonce or self._random_string(16),
            "iat": self._get_utc_epoch()
        }

        if username is not None:
            payload['username'] = username
        elif user_id is not None:
            payload['user_id'] = user_id

        jwt_headers = {"alg": algo, "typ": "JWT"}

        try:
            return jwt.encode(
                payload, self._private_key, algorithm=algo,
                headers=jwt_headers
            )
        except Exception as e:
            raise LoginIDError(500, "SERVER_ERROR",
                               f'Failed to generate service token {e}')



from optparse import Option
from typing import Optional
from . import _LoginIdClient, LoginIDError


class _Credentials:
    def init_add_fido2_credential_with_code(self: _LoginIdClient,
                                            code: str, code_type: str, username: str,
                                            display_name: Optional[str] = None,
                                            roaming_authenticator: bool = False) -> dict:
        """initialize adding a FIDO2 credential with pre-authorized code

        Args:
            code (str): The authorization code
            code_type (str): The code type, must be listed in `ALLOWED_CODE_TYPES`
            username (str): The username to add the new credential.
            display_name (str, optional): A human-palatable name for the user account, intended only for display. Defaults to None.
            roaming_authenticator (bool, optional): Whether the credential is a security key. Defaults to False.

        Returns:
            dict: the FIDO2 attestation payload.
        """
        url = '/credentials/fido2/init/code'

        payload = {
            "client_id": self._client_id,
            "authentication_code": {
                "code": code,
                "type": code_type
            },
        }

        options = {}
        if display_name:
            options["display_name"] = display_name
        if roaming_authenticator:
            options["roaming_authenticator"] = True
        if options:
            payload["options"] = options

        self._update_payload(payload, username=username)

        return self._request('post', url, "credentials.add", payload)

    def complete_add_fido2_credential(self: _LoginIdClient, attestation_payload: dict,
                                      username: str, credential_name: Optional[str] = None) -> dict:
        """Complete adding a FIDO2 credential (initialized with or without code)

        Args:
            attestation_payload (dict): The attestation payload returned by init functions
            username (str): The username to add the new credential.
            credential_name (str, optional): The name of the new credential. Defaults to None.

        Returns:
            dict: The attestation payload for the credentials if no errors

        Raises:
            LoginIDError: If there is an error
        """
        url = '/credentials/fido2/complete'

        payload = {
            "client_id": self._client_id,
            "attestation_payload": attestation_payload
        }
        self._update_payload(payload, username=username)

        if credential_name:
            payload["options"] = {
                "credential_name": credential_name
            }

        attestation_response = self._request(
            'post', url, payload=payload)

        # validate the jwt
        token = attestation_response.get("jwt", "")
        if not self.verify_token(token):
            raise LoginIDError(401, "invalid_jwt", "Invalid JWT")

        return attestation_response



import jwt
from .core import LoginIDError, _LoginIdClient, _Auth, _Credentials

import abc
import base64
import hashlib

from typing import Optional


class LoginID(_Auth, _Credentials, _LoginIdClient):
    """ This server SDK leverages either a web or mobile application
    and requires an API credential to be assigned to that integration.

    Args:
        client_id (str): Client ID as per the value provided on the dashboard
        private_key (str, optional): Private key of the API credential if one is assigned to the application on the dashboard.
        base_url (str, optional): Base URL where requests should be made (if specifying which environment is being used). Defaults to the LoginID production URL if not given (`None`)

    Example::

        from loginid import LoginID
        lid = LoginID(CLIENT_ID, PRIVATE_KEY)
    """
    ###########################################################################
    # Core client APIs
    ###########################################################################
    @abc.abstractmethod
    def generate_service_token(self, scope: str, username: Optional[str] = None, user_id: Optional[str] = None, algo: Optional[str] = None, nonce: Optional[str] = None) -> str:
        return super().generate_service_token(scope, username, user_id, algo, nonce)

    @abc.abstractmethod
    def verify_token(self, token: str, username: Optional[str] = None) -> bool:
        return super().verify_token(token, username)

    @abc.abstractmethod
    def client_id(self) -> str:
        return super().client_id()

    ###########################################################################
    # Transaction APIs
    ###########################################################################
    def generate_tx_auth_token(self, tx_payload: str,
                               nonce: Optional[str] = None,
                               algo: Optional[str] = None) -> str:
        """Generate an Authorization Token for Transaction Flow

        Args:
            tx_payload (str): The transaction payload
            nonce (str, optional): optional nonce for the token, auto-generated if not given
            algo (str, optional): Encryption algorithm, defaults to `"ES256"`

        Returns:
            str: The JWT authorization token
        """

        algo = algo or self.DEFAULT_JWT_ALOGORITHM
        if algo not in self.SUPPORTED_JWT_ALGORITHMS:
            raise LoginIDError(400, "invalid_algorithm", f"{algo} is not an allowed algorithm.")

        # hash and encode tx_payload
        payload_hash = hashlib.sha256(tx_payload.encode()).digest()
        payload_hash = (base64.urlsafe_b64encode(payload_hash)
                              .decode().strip('=')
                        )

        payload = {
            'scope': 'tx.create',
            'nonce': nonce or self._random_string(16),
            'payload_hash': payload_hash,
            'iat': self._get_utc_epoch()
        }

        return jwt.encode(
            payload, self._private_key,
            algorithm=algo,
            headers={'alg': algo, 'typ': 'JWT'}
        )

    def create_tx(self, tx_payload: str, nonce: str = None) -> str:
        """Create a transaction and return its ID

        Args:
            tx_payload (str): The transaction payload
            nonce (str, optional): The optional nonce, randomly generated if not provided

        Returns:
            str: The transaction id if no error
        """

        headers = {
            "Content-type": "application/json",
            "Authorization": "Bearer " + self.generate_tx_auth_token(tx_payload)
        }
        payload = dict(
            client_id=self._client_id,
            tx_payload=tx_payload,
            nonce=nonce or self._random_string()
        )

        return self._request('post', "/tx", payload=payload, headers=headers)['id']

    def verify_transaction(self, tx_token: str, tx_payload: str) -> bool:
        """Verify the jwt token returned upon completion of a transaction

        Args:
            tx_token (str): The JWT token
            tx_payload (str): the original transaction payload

        Returns: `True` if the token is valid, `False` if not
        """
        headers = jwt.get_unverified_header(tx_token)
        algo, kid = headers.get('alg'), headers.get('kid', '')
        if algo not in self.SUPPORTED_JWT_ALGORITHMS:
            raise ValueError(f"{algo} is not an allowed algorithm")

        # get public key
        public_key = self._get_public_key(kid)

        payload = jwt.decode(tx_token, public_key, algorithms=algo,
                             audience=self._client_id)

        to_hash = "".join([
            tx_payload,
            payload.get("nonce", ""),
            payload.get("server_nonce", "")
        ])
        hash = hashlib.sha256(to_hash.encode()).digest()
        hash = base64.urlsafe_b64encode(hash).decode().strip('=')

        return payload.get('tx_hash') == hash

    ###########################################################################
    # Auth APIs
    ###########################################################################
    # Fifo2
    @abc.abstractmethod
    def register_fido2_init(self, username: str,
                            display_name: Optional[str] = None,
                            roaming_authenticator: bool = False) -> dict:
        return super().register_fido2_init(username, display_name, roaming_authenticator)

    @abc.abstractmethod
    def register_fido2_complete(self, username: str, attestation_payload: str,
                                credential_name: Optional[str] = None) -> dict:
        return super().register_fido2_complete(username, attestation_payload, credential_name)

    @abc.abstractmethod
    def authenticate_fido2_init(self, username: str) -> dict:
        return super().authenticate_fido2_init(username)

    @abc.abstractmethod
    def authenticate_fido2_complete(self, username: str, assertion_payload: str) -> dict:
        return super().authenticate_fido2_complete(username, assertion_payload)

    # password
    @abc.abstractmethod
    def register_password(self, username: str, password: str) -> dict:
        return super().register_password(username, password)

    @abc.abstractmethod
    def authenticate_password(self, username: str, password: str) -> dict:
        return super().authenticate_password(username, password)

    # document scan
    @abc.abstractmethod
    def authenticate_doc_scan_init(self, username: str, credential_id: Optional[str] = None) -> dict:
        return super().authenticate_doc_scan_init(username, credential_id)

    @abc.abstractmethod
    def authenticate_doc_scan_complete(self, username: str, credential_id: str) -> dict:
        return super().authenticate_doc_scan_complete(username, credential_id)

    # publickey
    @abc.abstractmethod
    def authenticate_publickey_init(self, username: str, publickey: str, publickey_alg: Optional[str] = "ES256") -> dict:
        return super().authenticate_publickey_init(username, publickey, publickey_alg)

    @abc.abstractmethod
    def authenticate_publickey_complete(self, username: str, challenge_id: str, assertion: str) -> dict:
        return super().authenticate_publickey_complete(username, challenge_id, assertion)

    ###########################################################################
    # Credentials APIs
    ###########################################################################
    @abc.abstractmethod
    def init_add_fido2_credential_with_code(self: _LoginIdClient, code: str,
                                            codeType: str, username: str,
                                            display_name: Optional[str] = None,
                                            roaming_authenticator: bool = False) -> dict:
        return super().init_add_fido2_credential_with_code(code, codeType, username, display_name, roaming_authenticator)

    @abc.abstractmethod
    def complete_add_fido2_credential(self: _LoginIdClient,
                                      attestation_payload: dict,
                                      username: str,
                                      credential_name: Optional[str] = None) -> dict:
        return super().complete_add_fido2_credential(attestation_payload, username, credential_name)



import setuptools

with open("README.md", 'r', encoding="utf-8") as fh:
    long_description = fh.read()

INSTALL_REQUIRES = [
        "requests",
        "cryptography",
        "pyjwt"
    ]

setuptools.setup(
    name="loginid",
    version="0.1.3",
    author="Quang Hoang",
    author_email="quang@loginid.io",
    description="Interface SDK for LoginID infrastructure",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/loginid1/python-sdk",
    project_urls={
        "Bug Tracker": "https://github.com/loginid1/python-sdk/issues",
        "Documentation": "https://docs.loginid.io/Server-SDKs/Python/python-get-started"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES
)



from distutils.log import Log
from .. import LoginID


class _UserManagement():
    """
    User management class.
    """

    def get_user(self: LoginID, username: str) -> dict:
        """Get the user ID from username

        Args:
            username (str): The username

        Returns:
            dict: the user profile if no errors

        Raises:
            LoginIDError: If there is an error
        """
        url = "/manage/users/retrieve"

        payload = {
            "username": username
        }

        return self._request('post', url, "users.retrieve", payload)

    def delete_by_username(self: LoginID, username: str) -> str:
        """Delete a User by the username

        Args:
            username (str): the username to be deleted

        Returns:
            str: success message if user deleted successfully

        Raises:
            LoginIDError: If there is an error
        """
        url = "/manage/users/delete"

        payload = {
            "username": username
        }

        r = self._request('post', url, "users.delete", payload, expect=204)

        return f"User {username} deleted"

    def delete_by_user_id(self: LoginID, user_id: str) -> dict:
        """Delete a user by user Id

        Args:
            user_id (str): the user id to be deleted

        Returns:
            str: success message if user deleted successfully

        Raises:
            LoginIDError: If there is an error
        """
        url = f"/manage/users/{user_id}"

        self._request(
            'delete', url, token_scope="users.delete", expect=204)

        return f'user with id {user_id} deleted'

    def activate_user_by_id(self: LoginID, user_id: str) -> dict:
        """Activate a previously deactivated user

        Args:
            user_id (str): the user id to be activated

        Returns:
            dict: user profiles

        Raises:
            LoginIDError: If there is an error
        """
        url = f"/manage/users/{user_id}/activate"

        return self._request('put', url, token_scope="users.activate")

    def deactivate_user_by_id(self: LoginID, user_id: str) -> dict:
        """Deactivate a currently active user

        Args:
            user_id (str): the user id to be deactivated

        Returns:
            dict: user profiles if no error

        Raises:
            LoginIDError: If there is an error
        """
        url = f"/manage/users/{user_id}/deactivate"

        return self._request('put', url, token_scope="users.deactivate")

    def add_user_without_credentials(self: LoginID, username: str) -> dict:
        """Add a new user without credentials. The new user can create new credentials with recovery flow

        Args:
            username (str): The username of the new user

        Returns:
            dict: The new user's profile

        Raises:
            LoginIDError: If there is an error
        """
        url = f'/manage/users'

        payload = {
            "username": username
        }

        return self._request('post', url, "users.create", payload)



from optparse import Option
from os import supports_effective_ids
import re
from typing import Optional

import abc

from . import LoginID
from .core import LoginIDError

from .managements.user_mgmt import _UserManagement
from .managements.code_mgmt import _CodeManagement
from .managements.cred_mgmt import _CredentialManagement


class LoginIdManagement(_UserManagement, _CodeManagement, _CredentialManagement, LoginID):
    """This server SDK can be used with a management application and requires an API credential to be assigned to that integration. 
    All calls made from this SDK are intended to be backend-to-backend calls, as the operations are sensitive.

    Args:
        client_id (str): Client ID as per the value provided on the dashboard
        private_key (str): Private key of the API credential assigned to the application on the dashboard.
        base_url (str, optional): Base URL where requests should be made (if specifying which environment is being used). Defaults to the LoginID production URL if not given (`None`)

    Example::

        from loginid import LoginIdManagememnt
        lManagement = LoginIdManagement(CLIENT_ID, PRIVATE_KEY)
    """

    def __init__(self, client_id: str, private_key: str, base_url: Optional[str] = None) -> None:
        if not private_key:
            raise LoginIDError(400, "BAD_REQUEST", "Private key is required")

        super().__init__(client_id, private_key, base_url=base_url)

    # User Management
    @abc.abstractmethod
    def get_user(self, username: str) -> str:
        return super().get_user(username)

    @abc.abstractmethod
    def delete_by_username(self, username: str) -> str:
        return super().delete_by_username(username)

    @abc.abstractmethod
    def delete_by_user_id(self, user_id: str) -> dict:
        return super().delete_by_user_id(user_id)

    @abc.abstractmethod
    def activate_user_by_id(self, user_id: str) -> dict:
        return super().activate_user_by_id(user_id)

    @abc.abstractmethod
    def deactivate_user_by_id(self, user_id: str) -> dict:
        return super().deactivate_user_by_id(user_id)

    @abc.abstractmethod
    def add_user_without_credentials(self, username: str) -> dict:
        return super().add_user_without_credentials(username)

    # Code Management
    @abc.abstractmethod
    def generate_code(self, code_type: str, code_purpose: str, isAuthorized: bool, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        return super().generate_code(code_type, code_purpose, isAuthorized, user_id, username)

    @abc.abstractmethod
    def authorize_code(self, code: str, code_type: str, code_purpose: str, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        return super().authorize_code(code, code_type, code_purpose, user_id, username)

    @abc.abstractmethod
    def invalidate_all_codes(self, code_type: str, code_purpose: str, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        return super().invalidate_all_codes(code_type, code_purpose, user_id, username)

    # Credential Management
    @abc.abstractmethod
    def get_credentials(self, status: Optional[str] = None, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        return super().get_credentials(status, user_id, username)

    @abc.abstractmethod
    def rename_credential(self, cred_id: str, updated_name: str, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        return super().rename_credential(cred_id, updated_name, user_id, username)

    @abc.abstractmethod
    def revoke_credential(self, cred_id: str, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        return super().revoke_credential(cred_id, user_id, username)

    @abc.abstractmethod
    def force_fido2_credential_init(self: LoginID, user_id: str,
                                    display_name: Optional[str] = None,
                                    roaming_authenticator: bool = False) -> dict:
        return super().force_fido2_credential_init(user_id, display_name, roaming_authenticator)

    @abc.abstractmethod
    def add_doc_scan_credential_init(self: LoginID,
                                     user_id: Optional[str] = None,
                                     username: Optional[str] = None,
                                     credential_name: Optional[str] = None) -> dict:
        return super().add_doc_scan_credential_init(user_id, username, credential_name)

    @abc.abstractmethod
    def add_doc_scan_credential_complete(self: LoginID, credential_id: str, activate_credential: bool,
                                         user_id: Optional[str] = None, username: Optional[str] = None,
                                         ) -> dict:
        return super().add_doc_scan_credential_complete(credential_id, activate_credential, user_id, username)

    @abc.abstractmethod
    def evaluate_doc_scan_credential(self: LoginID, credential_id: str, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        return super().evaluate_doc_scan_credential(credential_id, user_id, username)

    @abc.abstractmethod
    def generate_recovery_code(self: LoginID, user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        return super().generate_recovery_code(user_id, username)

    @abc.abstractmethod
    def add_publickey_credential(self: LoginID, publickey: str, publickey_alg: Optional[str] = "ES256", user_id: Optional[str] = None, username: Optional[str] = None, credential_name: Optional[str] = None) -> dict:
        return super().add_publickey_credential(publickey, publickey_alg, user_id, username, credential_name)



from optparse import Option
from .. import LoginID
from typing import Optional


class _CodeManagement():
    """
    Code management class.
    """

    def generate_code(self: LoginID, code_type: str,
                      code_purpose: str, is_authorized: bool,
                      user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        """Generate a code

        Args:
            code_type (str): The code type
            code_purpose (str): The purpose of the code
            is_authorized (bool): Whether the code authorizes the user or not
            user_id (str, optional): The user id to generate the code for. Defaults to None.
            username (str, optional): The username, must be present if `user_id` is not . Defaults to None.

        Returns:
            dict: the response body from code generation request if no errors, else raise LoginIDError
        """
        self.validate_code_type(code_type)

        url = f"/codes/{code_type}/generate"

        payload = dict(
            client_id=self._client_id,
            purpose=code_purpose,
            authorize=is_authorized
        )
        self._update_payload(payload, user_id, username)

        return self._request('post', url, "codes.generate", payload=payload)

    def authorize_code(self: LoginID, code: str,
                       code_type: str, code_purpose: str,
                       user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        """Authorize a given code

        Args:
            code (str): The code that needs authorization
            code_type (str): The type of the code
            code_purpose (str): The purpose of the code
            user_id (str, optional): The user_id associate with the code. Defaults to None.
            username (str, optional): The username, must be present if `user_id` is not . Defaults to None.

        Returns:
            dict: The response body from code authorization
        """
        self.validate_code_type(code_type)

        url = f"/codes/{code_type}/authorize"

        payload = dict(
            client_id=self._client_id,
            purpose=code_purpose,
            code=code
        )
        self._update_payload(payload, user_id, username)

        return self._request('post', url, "codes.authorize", payload=payload)

    def invalidate_all_codes(self: LoginID,
                             code_type: str, code_purpose: str,
                             user_id: Optional[str] = None, username: Optional[str] = None) -> dict:
        """Invalidate all codes of a given type

        Args:
            code_type (str): code type
            code_purpose (str): the purpose of the code
            user_id (str, optional): The user_id associate with the code. Defaults to None.
            username (str, optional): The username, must be present if `user_id` is not . Defaults to None.

        Returns:
            dict: The response body from code invalidation::

                {
                    "deleted_at": "2020-04-01T00:00:00.000Z",
                }

        Raises:
            LoginIDError: If there is an error
        """
        self.validate_code_type(code_type)

        url = f'/codes/{code_type}/invalidate-all'

        payload = dict(
            client_id=self._client_id,
            purpose=code_purpose,
        )
        self._update_payload(payload, user_id, username)

        return self._request('post', url, "codes.invalidate", payload=payload)
